use super::Gte;

#[test]
fn gte_lzcr() {
    let expected = [
        (0x00000000, 32),
        (0xffffffff, 32),
        (0x00000001, 31),
        (0x80000000, 1),
        (0x7fffffff, 1),
        (0xdeadbeef, 2),
        (0x000c0ffe, 12),
        (0xfffc0ffe, 14),
    ];

    let mut gte = Gte::new();

    for &(lzcs, lzcr) in &expected {
        gte.set_data(30, lzcs);

        let r = gte.data(31);

        assert!(r == lzcr);
    }
}

#[test]
fn gte_ops() {
    for test in TESTS {
        println!("Test: '{}'", test.desc);
        println!("Command: 0x{:08x}", test.command);

        let mut gte = test.initial.make_gte();

        gte.command(test.command);

        test.result.validate(gte);
    }
}

struct Test {
    /// Test description
    desc: &'static str,
    /// Initial GTE configuration
    initial: Config,
    /// GTE command being executed
    command: u32,
    /// GTE configuration post-command
    result: Config,
}

/// GTE register config: slice of couples `(register_offset,
/// register_value)`. Missing registers are set to 0.
struct Config {
    /// Control registers
    controls: &'static [(u8, u32)],
    /// Data registers
    data: &'static [(u8, u32)],
}

impl Config {
    fn make_gte(&self) -> Gte {
        let mut gte = Gte::new();

        for &(reg, val) in self.controls {
            gte.set_control(reg as u32, val);
        }

        for &(reg, val) in self.data {
            if reg == 15 {
                // Writing to 14 should set this register and writing
                // here will push a new entry onto the XY_FIFO which
                // will change the previous values.
                continue;
            }

            if reg == 28 {
                // This sets the IR1...3 registers MSB but those
                // values should have been set through registers
                // 9...11
                continue;
            }

            if reg == 29 {
                // This register is read only
                continue;
            }

            gte.set_data(reg as u32, val);
        }

        gte
    }

    fn validate(&self, gte: Gte) {
        let mut error_count = 0u32;

        for &(reg, val) in self.controls {
            let v = gte.control(reg as u32);

            if v != val {
                println!("Control register {}: expected 0x{:08x} got 0x{:08x}",
                         reg, val, v);
                error_count += 1;
            }
        }

        for &(reg, val) in self.data {
            let v = gte.data(reg as u32);

            if v != val {
                println!("Data register {}: expected 0x{:08x} got 0x{:08x}",
                         reg, val, v);
                error_count += 1;
            }
        }

        if error_count > 0 {
            panic!("{} registers errors", error_count);
        }
    }
}

/// Reference data generated using tests/gte_commands/main.s in
/// https://github.com/simias/psx-hardware-tests and running it on the
/// real console.
static TESTS: &'static [Test] = &[
    Test {
        desc: "GTE_RTPT, lm=0, cv=0, v=0, mx=0, sf=1",
        initial: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                ],
            data: &[
                (0, 0x00e70119),
                (1, 0xfffffe65),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (31, 0x00000020),
                ],
        },
        command: 0x00080030,
        result: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                (31, 0x00001000),
                ],
            data: &[
                (0, 0x00e70119),
                (1, 0xfffffe65),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (8, 0x00001000),
                (9, 0x0000012b),
                (10, 0xfffffff0),
                (11, 0x000015d9),
                (12, 0x00f40176),
                (13, 0x00f9016b),
                (14, 0x00ed0176),
                (15, 0x00ed0176),
                (17, 0x000015eb),
                (18, 0x000015aa),
                (19, 0x000015d9),
                (24, 0x0106e038),
                (25, 0x0000012b),
                (26, 0xfffffff0),
                (27, 0x000015d9),
                (28, 0x00007c02),
                (29, 0x00007c02),
                (31, 0x00000020),
                ],
        },
    },

    Test {
        desc: "GTE_NCLIP, lm=0, cv=0, v=0, mx=0, sf=0",
        initial: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                (31, 0x00001000),
                ],
            data: &[
                (0, 0x00e70119),
                (1, 0xfffffe65),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (8, 0x00001000),
                (9, 0x0000012b),
                (10, 0xfffffff0),
                (11, 0x000015d9),
                (12, 0x00f40176),
                (13, 0x00f9016b),
                (14, 0x00ed0176),
                (15, 0x00ed0176),
                (17, 0x000015eb),
                (18, 0x000015aa),
                (19, 0x000015d9),
                (24, 0x0106e038),
                (25, 0x0000012b),
                (26, 0xfffffff0),
                (27, 0x000015d9),
                (28, 0x00007c02),
                (29, 0x00007c02),
                (31, 0x00000020),
                ],
        },
        command: 0x00000006,
        result: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                ],
            data: &[
                (0, 0x00e70119),
                (1, 0xfffffe65),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (8, 0x00001000),
                (9, 0x0000012b),
                (10, 0xfffffff0),
                (11, 0x000015d9),
                (12, 0x00f40176),
                (13, 0x00f9016b),
                (14, 0x00ed0176),
                (15, 0x00ed0176),
                (17, 0x000015eb),
                (18, 0x000015aa),
                (19, 0x000015d9),
                (24, 0x0000004d),
                (25, 0x0000012b),
                (26, 0xfffffff0),
                (27, 0x000015d9),
                (28, 0x00007c02),
                (29, 0x00007c02),
                (31, 0x00000020),
                ],
        },
    },

    Test {
        desc: "GTE_AVSZ3, lm=0, cv=0, v=0, mx=0, sf=1",
        initial: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                ],
            data: &[
                (0, 0x00e70119),
                (1, 0xfffffe65),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (8, 0x00001000),
                (9, 0x0000012b),
                (10, 0xfffffff0),
                (11, 0x000015d9),
                (12, 0x00f40176),
                (13, 0x00f9016b),
                (14, 0x00ed0176),
                (15, 0x00ed0176),
                (17, 0x000015eb),
                (18, 0x000015aa),
                (19, 0x000015d9),
                (24, 0x0000004d),
                (25, 0x0000012b),
                (26, 0xfffffff0),
                (27, 0x000015d9),
                (28, 0x00007c02),
                (29, 0x00007c02),
                (31, 0x00000020),
                ],
        },
        command: 0x0008002d,
        result: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                ],
            data: &[
                (0, 0x00e70119),
                (1, 0xfffffe65),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (7, 0x00000572),
                (8, 0x00001000),
                (9, 0x0000012b),
                (10, 0xfffffff0),
                (11, 0x000015d9),
                (12, 0x00f40176),
                (13, 0x00f9016b),
                (14, 0x00ed0176),
                (15, 0x00ed0176),
                (17, 0x000015eb),
                (18, 0x000015aa),
                (19, 0x000015d9),
                (24, 0x00572786),
                (25, 0x0000012b),
                (26, 0xfffffff0),
                (27, 0x000015d9),
                (28, 0x00007c02),
                (29, 0x00007c02),
                (31, 0x00000020),
                ],
        },
    },

    Test {
        desc: "GTE_NCDS, lm=1, cv=0, v=0, mx=0, sf=1",
        initial: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                ],
            data: &[
                (0, 0x00000b50),
                (1, 0xfffff4b0),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (6, 0x2094a539),
                (7, 0x00000572),
                (8, 0x00001000),
                (9, 0x0000012b),
                (10, 0xfffffff0),
                (11, 0x000015d9),
                (12, 0x00f40176),
                (13, 0x00f9016b),
                (14, 0x00ed0176),
                (15, 0x00ed0176),
                (17, 0x000015eb),
                (18, 0x000015aa),
                (19, 0x000015d9),
                (24, 0x00572786),
                (25, 0x0000012b),
                (26, 0xfffffff0),
                (27, 0x000015d9),
                (28, 0x00007c02),
                (29, 0x00007c02),
                (31, 0x00000020),
                ],
        },
        command: 0x00080413,
        result: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                (31, 0x81f00000),
                ],
            data: &[
                (0, 0x00000b50),
                (1, 0xfffff4b0),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (6, 0x2094a539),
                (7, 0x00000572),
                (8, 0x00001000),
                (12, 0x00f40176),
                (13, 0x00f9016b),
                (14, 0x00ed0176),
                (15, 0x00ed0176),
                (17, 0x000015eb),
                (18, 0x000015aa),
                (19, 0x000015d9),
                (22, 0x20000000),
                (24, 0x00572786),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (31, 0x00000020),
                ],
        },
    },
    Test {
        desc: "GTE_DCPS, lm=0, cv=0, v=0, mx=0, sf=1",
        initial: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                ],
            data: &[
                (0, 0x00000b50),
                (1, 0xfffff4b0),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (6, 0x2094a539),
                (7, 0x00000572),
                (8, 0x00001000),
                (9, 0x0000012b),
                (10, 0xfffffff0),
                (11, 0x000015d9),
                (12, 0x00f40176),
                (13, 0x00f9016b),
                (14, 0x00ed0176),
                (15, 0x00ed0176),
                (17, 0x000015eb),
                (18, 0x000015aa),
                (19, 0x000015d9),
                (24, 0x00572786),
                (25, 0x0000012b),
                (26, 0xfffffff0),
                (27, 0x000015d9),
                (28, 0x00007c02),
                (29, 0x00007c02),
                (31, 0x00000020),
                ],
        },
        command: 0x00080010,
        result: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                ],
            data: &[
                (0, 0x00000b50),
                (1, 0xfffff4b0),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (6, 0x2094a539),
                (7, 0x00000572),
                (8, 0x00001000),
                (12, 0x00f40176),
                (13, 0x00f9016b),
                (14, 0x00ed0176),
                (15, 0x00ed0176),
                (17, 0x000015eb),
                (18, 0x000015aa),
                (19, 0x000015d9),
                (22, 0x20000000),
                (24, 0x00572786),
                (31, 0x00000020),
                ],
        },
    },
    Test {
        desc: "GTE_RTPS, lm=0, cv=0, v=0, mx=0, sf=1",
        initial: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                ],
            data: &[
                (0, 0x00000b50),
                (1, 0xfffff4b0),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (6, 0x2094a539),
                (8, 0x00001000),
                (31, 0x00000020),
                ],
        },
        command: 0x00080001,
        result: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                (31, 0x80004000),
                ],
            data: &[
                (0, 0x00000b50),
                (1, 0xfffff4b0),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (6, 0x2094a539),
                (8, 0x00000e08),
                (9, 0x00000bd1),
                (10, 0x000002dc),
                (11, 0x00000d12),
                (14, 0x01d003ff),
                (15, 0x01d003ff),
                (19, 0x00000d12),
                (24, 0x00e08388),
                (25, 0x00000bd1),
                (26, 0x000002dc),
                (27, 0x00000d12),
                (28, 0x000068b7),
                (29, 0x000068b7),
                (31, 0x00000020),
                ],
        },
    },
    Test {
        desc: "GTE_NCCT, lm=0, cv=0, v=0, mx=0, sf=1",
        initial: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                ],
            data: &[
                (0, 0x00000b50),
                (1, 0xfffff4b0),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (6, 0x2094a539),
                (7, 0x00000572),
                (8, 0x00001000),
                (12, 0x00f40176),
                (13, 0x00f9016b),
                (14, 0x00ed0176),
                (15, 0x00ed0176),
                (17, 0x000015eb),
                (18, 0x000015aa),
                (19, 0x000015d9),
                (24, 0x00572786),
                (31, 0x00000020),
                ],
        },
        command: 0x0008003f,
        result: Config {
            controls: &[
                (0, 0x00000ffb),
                (1, 0xffb7ff44),
                (2, 0xf9ca0ebc),
                (3, 0x063700ad),
                (4, 0x00000eb7),
                (6, 0xfffffeac),
                (7, 0x00001700),
                (9, 0x00000fa0),
                (10, 0x0000f060),
                (11, 0x0000f060),
                (13, 0x00000640),
                (14, 0x00000640),
                (15, 0x00000640),
                (16, 0x0bb80fa0),
                (17, 0x0fa00fa0),
                (18, 0x0fa00bb8),
                (19, 0x0bb80fa0),
                (20, 0x00000fa0),
                (24, 0x01400000),
                (25, 0x00f00000),
                (26, 0x00000400),
                (27, 0xfffffec8),
                (28, 0x01400000),
                (29, 0x00000155),
                (30, 0x00000100),
                (31, 0x00380000),
                ],
            data: &[
                (0, 0x00000b50),
                (1, 0xfffff4b0),
                (2, 0x00e700d5),
                (3, 0xfffffe21),
                (4, 0x00b90119),
                (5, 0xfffffe65),
                (6, 0x2094a539),
                (7, 0x00000572),
                (8, 0x00001000),
                (9, 0x000000b3),
                (10, 0x00000207),
                (11, 0x000001d1),
                (12, 0x00f40176),
                (13, 0x00f9016b),
                (14, 0x00ed0176),
                (15, 0x00ed0176),
                (17, 0x000015eb),
                (18, 0x000015aa),
                (19, 0x000015d9),
                (20, 0x20000000),
                (21, 0x201b1f0a),
                (22, 0x201d200b),
                (24, 0x00572786),
                (25, 0x000000b3),
                (26, 0x00000207),
                (27, 0x000001d1),
                (28, 0x00000c81),
                (29, 0x00000c81),
                (31, 0x00000020),
                ],
        },
    },

    Test {
        desc: "GTE_RTPT, lm=0, cv=0, v=0, mx=0, sf=1 full 0xffffffff",
        initial: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0xfffff000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0x00000f80),
                (10, 0x00000f80),
                (11, 0x00000f80),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
        command: 0x00080030,
        result: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0x80061000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xfffefffe),
                (13, 0xfffefffe),
                (14, 0xfffefffe),
                (15, 0xfffefffe),
                (16, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xfffe0000),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
    },
    Test {
        desc: "GTE_RTPS, lm=0, cv=0, v=0, mx=0, sf=1 full 0xffffffff",
        initial: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0xfffff000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0x00000f80),
                (10, 0x00000f80),
                (11, 0x00000f80),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
        command: 0x00080001,
        result: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0x80061000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xfffefffe),
                (15, 0xfffefffe),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xfffe0000),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
    },
    Test {
        desc: "GTE_NCLIP, lm=0, cv=0, v=0, mx=0, sf=1 full 0xffffffff",
        initial: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0xfffff000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0x00000f80),
                (10, 0x00000f80),
                (11, 0x00000f80),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
        command: 0x00080006,
        result: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0x00000f80),
                (10, 0x00000f80),
                (11, 0x00000f80),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
    },
    Test {
        desc: "GTE_DPCS, lm=0, cv=0, v=0, mx=0, sf=1 full 0xffffffff",
        initial: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0xfffff000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0x00000f80),
                (10, 0x00000f80),
                (11, 0x00000f80),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
        command: 0x00080010,
        result: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0x00000ff0),
                (10, 0x00000ff0),
                (11, 0x00000ff0),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0x00000ff0),
                (26, 0x00000ff0),
                (27, 0x00000ff0),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
    },
    Test {
        desc: "GTE_NCDS, lm=0, cv=0, v=0, mx=0, sf=1 full 0xffffffff",
        initial: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0xfffff000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0x00000f80),
                (10, 0x00000f80),
                (11, 0x00000f80),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
        command: 0x00080013,
        result: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0x00380000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xff000000),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
    },
    Test {
        desc: "GTE_NCCT, lm=0, cv=0, v=0, mx=0, sf=1 full 0xffffffff",
        initial: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0xfffff000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0x00000f80),
                (10, 0x00000f80),
                (11, 0x00000f80),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
        command: 0x0008003f,
        result: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0x00380000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xff000000),
                (21, 0xff000000),
                (22, 0xff000000),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
    },
    Test {
        desc: "GTE_AVSZ3, lm=0, cv=0, v=0, mx=0, sf=1 full 0xffffffff",
        initial: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0xfffff000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0x00000f80),
                (10, 0x00000f80),
                (11, 0x00000f80),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
        command: 0x0008002d,
        result: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0x80040000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (8, 0xffffffff),
                (9, 0x00000f80),
                (10, 0x00000f80),
                (11, 0x00000f80),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xfffd0003),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
    },
    Test {
        desc: "GTE_MVMVA, lm=0, cv=0, v=0, mx=0, sf=1 full 0xffffffff",
        initial: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0xfffff000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0x00000f80),
                (10, 0x00000f80),
                (11, 0x00000f80),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
        command: 0x00080012,
        result: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
    },
    Test {
        desc: "GTE_MVMVA, lm=0, cv=3, v=3, mx=0, sf=1 full 0xffffffff",
        initial: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0xfffff000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0x00000f80),
                (10, 0x00000f80),
                (11, 0x00000f80),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
        command: 0x0009e012,
        result: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0xfffffffd),
                (10, 0xfffffffd),
                (11, 0xfffffffd),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xfffffffd),
                (26, 0xfffffffd),
                (27, 0xfffffffd),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
    },
    Test {
        desc: "GTE_OP GTE_DCPL, lm=0, cv=0, v=0, mx=0, sf=1 full 0xffffffff",
        initial: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                (31, 0xfffff000),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0x00000f80),
                (10, 0x00000f80),
                (11, 0x00000f80),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
        command: 0x00080029,
        result: Config {
            controls: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0xffffffff),
                (8, 0xffffffff),
                (9, 0xffffffff),
                (10, 0xffffffff),
                (11, 0xffffffff),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0xffffffff),
                (17, 0xffffffff),
                (18, 0xffffffff),
                (19, 0xffffffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xffffffff),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0xffffffff),
                (26, 0xffffffff),
                (27, 0xffffffff),
                (28, 0xffffffff),
                (29, 0xffffffff),
                (30, 0xffffffff),
                ],
            data: &[
                (0, 0xffffffff),
                (1, 0xffffffff),
                (2, 0xffffffff),
                (3, 0xffffffff),
                (4, 0xffffffff),
                (5, 0xffffffff),
                (6, 0xffffffff),
                (7, 0x0000ffff),
                (8, 0xffffffff),
                (9, 0x00000f71),
                (10, 0x00000f71),
                (11, 0x00000f71),
                (12, 0xffffffff),
                (13, 0xffffffff),
                (14, 0xffffffff),
                (15, 0xffffffff),
                (16, 0x0000ffff),
                (17, 0x0000ffff),
                (18, 0x0000ffff),
                (19, 0x0000ffff),
                (20, 0xffffffff),
                (21, 0xffffffff),
                (22, 0xfff7f7f7),
                (23, 0xffffffff),
                (24, 0xffffffff),
                (25, 0x00000f71),
                (26, 0x00000f71),
                (27, 0x00000f71),
                (28, 0x00007bde),
                (29, 0x00007bde),
                (30, 0xffffffff),
                (31, 0x00000020),
                ],
        },
    },
    Test {
        desc: "GTE_OP GTE_MVMVA, lm=0, cv=0, v=0, mx=0, sf=1 random",
        initial: Config {
            controls: &[
                (0, 0xff35cdf4),
                (1, 0xf8acd6a6),
                (2, 0x1954aa70),
                (3, 0xae7b5062),
                (4, 0x00000c63),
                (5, 0xcad4cc39),
                (6, 0xb9c11958),
                (7, 0xa942b312),
                (8, 0xaf436779),
                (9, 0x3c2d507a),
                (10, 0x95f99741),
                (11, 0x72413224),
                (12, 0x0000499d),
                (13, 0x0a37d280),
                (14, 0xdbe8feec),
                (15, 0x2395909a),
                (16, 0x47364c98),
                (17, 0x795c2ed7),
                (18, 0x637e48f4),
                (19, 0x89557da5),
                (20, 0xffff997a),
                (21, 0x690eb551),
                (22, 0x3dfb368e),
                (23, 0x2bbe355f),
                (24, 0xb07c9d22),
                (25, 0x030c876b),
                (26, 0x00003b7d),
                (27, 0x0000765a),
                (28, 0x228c2901),
                (29, 0xffffe86f),
                (30, 0xffffaf93),
                (31, 0xc741f000),
                ],
            data: &[
                (0, 0x91d5c574),
                (1, 0xffffdf9c),
                (2, 0xcea213bc),
                (3, 0x0000143e),
                (4, 0x2360a947),
                (5, 0x00003248),
                (6, 0x1747e72e),
                (7, 0x0000cc08),
                (8, 0x0000381d),
                (9, 0xffffe2ff),
                (10, 0xffffe0f8),
                (11, 0xffffe1b6),
                (12, 0x9da7438d),
                (13, 0xff60f0ed),
                (14, 0xbf5961ab),
                (15, 0xbf5961ab),
                (16, 0x0000b1c1),
                (17, 0x0000dda6),
                (18, 0x0000ce75),
                (19, 0x0000b2d1),
                (20, 0xdb01b77a),
                (21, 0x19cd28cd),
                (22, 0x1a75d97a),
                (23, 0xe91dc0ad),
                (24, 0x764e464f),
                (25, 0x4aa5a1e5),
                (26, 0x3b1a1977),
                (27, 0x39fb3f5f),
                (30, 0xfe8de0c9),
                (31, 0x00000007),
                ],
        },
        command: 0x00080012,
        result: Config {
            controls: &[
                (0, 0xff35cdf4),
                (1, 0xf8acd6a6),
                (2, 0x1954aa70),
                (3, 0xae7b5062),
                (4, 0x00000c63),
                (5, 0xcad4cc39),
                (6, 0xb9c11958),
                (7, 0xa942b312),
                (8, 0xaf436779),
                (9, 0x3c2d507a),
                (10, 0x95f99741),
                (11, 0x72413224),
                (12, 0x0000499d),
                (13, 0x0a37d280),
                (14, 0xdbe8feec),
                (15, 0x2395909a),
                (16, 0x47364c98),
                (17, 0x795c2ed7),
                (18, 0x637e48f4),
                (19, 0x89557da5),
                (20, 0xffff997a),
                (21, 0x690eb551),
                (22, 0x3dfb368e),
                (23, 0x2bbe355f),
                (24, 0xb07c9d22),
                (25, 0x030c876b),
                (26, 0x00003b7d),
                (27, 0x0000765a),
                (28, 0x228c2901),
                (29, 0xffffe86f),
                (30, 0xffffaf93),
                (31, 0x81c00000),
                ],
            data: &[
                (0, 0x91d5c574),
                (1, 0xffffdf9c),
                (2, 0xcea213bc),
                (3, 0x0000143e),
                (4, 0x2360a947),
                (5, 0x00003248),
                (6, 0x1747e72e),
                (7, 0x0000cc08),
                (8, 0x0000381d),
                (9, 0xffff8000),
                (10, 0xffff8000),
                (11, 0xffff8000),
                (12, 0x9da7438d),
                (13, 0xff60f0ed),
                (14, 0xbf5961ab),
                (15, 0xbf5961ab),
                (16, 0x0000b1c1),
                (17, 0x0000dda6),
                (18, 0x0000ce75),
                (19, 0x0000b2d1),
                (20, 0xdb01b77a),
                (21, 0x19cd28cd),
                (22, 0x1a75d97a),
                (23, 0xe91dc0ad),
                (24, 0x764e464f),
                (25, 0xcad5dc86),
                (26, 0xb9c34e06),
                (27, 0xa943a529),
                (30, 0xfe8de0c9),
                (31, 0x00000007),
                ],
        },
    },
    Test {
        desc: "GTE_OP GTE_RTPS, lm=0, cv=0, v=0, mx=0, sf=1 random",
        initial: Config {
            controls: &[
                (0, 0xff35cdf4),
                (1, 0xf8acd6a6),
                (2, 0x1954aa70),
                (3, 0xae7b5062),
                (4, 0x00000c63),
                (5, 0xcad4cc39),
                (6, 0xb9c11958),
                (7, 0xa942b312),
                (8, 0xaf436779),
                (9, 0x3c2d507a),
                (10, 0x95f99741),
                (11, 0x72413224),
                (12, 0x0000499d),
                (13, 0x0a37d280),
                (14, 0xdbe8feec),
                (15, 0x2395909a),
                (16, 0x47364c98),
                (17, 0x795c2ed7),
                (18, 0x637e48f4),
                (19, 0x89557da5),
                (20, 0xffff997a),
                (21, 0x690eb551),
                (22, 0x3dfb368e),
                (23, 0x2bbe355f),
                (24, 0x307c9d22),
                (25, 0x030c876b),
                (26, 0x00003b7d),
                (27, 0x0000765a),
                (28, 0x228c2901),
                (29, 0xffffe86f),
                (30, 0xffffaf93),
                (31, 0xc741f000),
                ],
            data: &[
                (0, 0x91d5c574),
                (1, 0xffffdf9c),
                (2, 0xcea213bc),
                (3, 0x0000143e),
                (4, 0x2360a947),
                (5, 0x00003248),
                (6, 0x1747e72e),
                (7, 0x0000cc08),
                (8, 0x0000381d),
                (9, 0xffffe2ff),
                (10, 0xffffe0f8),
                (11, 0xffffe1b6),
                (12, 0x9da7438d),
                (13, 0xff60f0ed),
                (14, 0xbf5961ab),
                (15, 0xbf5961ab),
                (16, 0x0000b1c1),
                (17, 0x0000dda6),
                (18, 0x0000ce75),
                (19, 0x0000b2d1),
                (20, 0xdb01b77a),
                (21, 0x19cd28cd),
                (22, 0x1a75d97a),
                (23, 0xe91dc0ad),
                (24, 0x764e464f),
                (25, 0x4aa5a1e5),
                (26, 0x3b1a1977),
                (27, 0x39fb3f5f),
                (30, 0xfe8de0c9),
                (31, 0x00000007),
                ],
        },
        command: 0x00080001,
        result: Config {
            controls: &[
                (0, 0xff35cdf4),
                (1, 0xf8acd6a6),
                (2, 0x1954aa70),
                (3, 0xae7b5062),
                (4, 0x00000c63),
                (5, 0xcad4cc39),
                (6, 0xb9c11958),
                (7, 0xa942b312),
                (8, 0xaf436779),
                (9, 0x3c2d507a),
                (10, 0x95f99741),
                (11, 0x72413224),
                (12, 0x0000499d),
                (13, 0x0a37d280),
                (14, 0xdbe8feec),
                (15, 0x2395909a),
                (16, 0x47364c98),
                (17, 0x795c2ed7),
                (18, 0x637e48f4),
                (19, 0x89557da5),
                (20, 0xffff997a),
                (21, 0x690eb551),
                (22, 0x3dfb368e),
                (23, 0x2bbe355f),
                (24, 0x307c9d22),
                (25, 0x030c876b),
                (26, 0x00003b7d),
                (27, 0x0000765a),
                (28, 0x228c2901),
                (29, 0xffffe86f),
                (30, 0xffffaf93),
                (31, 0x81c7f000),
                ],
            data: &[
                (0, 0x91d5c574),
                (1, 0xffffdf9c),
                (2, 0xcea213bc),
                (3, 0x0000143e),
                (4, 0x2360a947),
                (5, 0x00003248),
                (6, 0x1747e72e),
                (7, 0x0000cc08),
                (8, 0x00001000),
                (9, 0xffff8000),
                (10, 0xffff8000),
                (11, 0xffff8000),
                (12, 0xff60f0ed),
                (13, 0xbf5961ab),
                (14, 0xfc00fc00),
                (15, 0xfc00fc00),
                (16, 0x0000dda6),
                (17, 0x0000ce75),
                (18, 0x0000b2d1),
                (20, 0xdb01b77a),
                (21, 0x19cd28cd),
                (22, 0x1a75d97a),
                (23, 0xe91dc0ad),
                (24, 0x0f3fb2a7),
                (25, 0xcad5dc86),
                (26, 0xb9c34e06),
                (27, 0xa943a529),
                (30, 0xfe8de0c9),
                (31, 0x00000007),
                ],
        },
    },
    Test {
        desc: "GTE_OP GTE_RTPT, lm=0, cv=0, v=0, mx=0, sf=1 random",
        initial: Config {
            controls: &[
                (0, 0xff35cdf4),
                (1, 0xf8acd6a6),
                (2, 0x1954aa70),
                (3, 0xae7b5062),
                (4, 0x00000c63),
                (5, 0xcad4cc39),
                (6, 0xb9c11958),
                (7, 0xa942b312),
                (8, 0xaf436779),
                (9, 0x3c2d507a),
                (10, 0x95f99741),
                (11, 0x72413224),
                (12, 0x0000499d),
                (13, 0x0a37d280),
                (14, 0xdbe8feec),
                (15, 0x2395909a),
                (16, 0x47364c98),
                (17, 0x795c2ed7),
                (18, 0x637e48f4),
                (19, 0x89557da5),
                (20, 0xffff997a),
                (21, 0x690eb551),
                (22, 0x3dfb368e),
                (23, 0x2bbe355f),
                (24, 0x307c9d22),
                (25, 0x030c876b),
                (26, 0x00003b7d),
                (27, 0x0000765a),
                (28, 0x228c2901),
                (29, 0xffffe86f),
                (30, 0xffffaf93),
                (31, 0xc741f000),
                ],
            data: &[
                (0, 0x91d5c574),
                (1, 0xffffdf9c),
                (2, 0xcea213bc),
                (3, 0x0000143e),
                (4, 0x2360a947),
                (5, 0x00003248),
                (6, 0x1747e72e),
                (7, 0x0000cc08),
                (8, 0x0000381d),
                (9, 0xffffe2ff),
                (10, 0xffffe0f8),
                (11, 0xffffe1b6),
                (12, 0x9da7438d),
                (13, 0xff60f0ed),
                (14, 0xbf5961ab),
                (15, 0xbf5961ab),
                (16, 0x0000b1c1),
                (17, 0x0000dda6),
                (18, 0x0000ce75),
                (19, 0x0000b2d1),
                (20, 0xdb01b77a),
                (21, 0x19cd28cd),
                (22, 0x1a75d97a),
                (23, 0xe91dc0ad),
                (24, 0x764e464f),
                (25, 0x4aa5a1e5),
                (26, 0x3b1a1977),
                (27, 0x39fb3f5f),
                (30, 0xfe8de0c9),
                (31, 0x00000007),
                ],
        },
        command: 0x00080030,
        result: Config {
            controls: &[
                (0, 0xff35cdf4),
                (1, 0xf8acd6a6),
                (2, 0x1954aa70),
                (3, 0xae7b5062),
                (4, 0x00000c63),
                (5, 0xcad4cc39),
                (6, 0xb9c11958),
                (7, 0xa942b312),
                (8, 0xaf436779),
                (9, 0x3c2d507a),
                (10, 0x95f99741),
                (11, 0x72413224),
                (12, 0x0000499d),
                (13, 0x0a37d280),
                (14, 0xdbe8feec),
                (15, 0x2395909a),
                (16, 0x47364c98),
                (17, 0x795c2ed7),
                (18, 0x637e48f4),
                (19, 0x89557da5),
                (20, 0xffff997a),
                (21, 0x690eb551),
                (22, 0x3dfb368e),
                (23, 0x2bbe355f),
                (24, 0x307c9d22),
                (25, 0x030c876b),
                (26, 0x00003b7d),
                (27, 0x0000765a),
                (28, 0x228c2901),
                (29, 0xffffe86f),
                (30, 0xffffaf93),
                (31, 0x81c7f000),
                ],
            data: &[
                (0, 0x91d5c574),
                (1, 0xffffdf9c),
                (2, 0xcea213bc),
                (3, 0x0000143e),
                (4, 0x2360a947),
                (5, 0x00003248),
                (6, 0x1747e72e),
                (7, 0x0000cc08),
                (8, 0x00001000),
                (9, 0xffff8000),
                (10, 0xffff8000),
                (11, 0xffff8000),
                (12, 0xfc00fc00),
                (13, 0xfc00fc00),
                (14, 0xfc00fc00),
                (15, 0xfc00fc00),
                (16, 0x0000b2d1),
                (20, 0xdb01b77a),
                (21, 0x19cd28cd),
                (22, 0x1a75d97a),
                (23, 0xe91dc0ad),
                (24, 0x0f3fb2a7),
                (25, 0xcad557c8),
                (26, 0xb9c0d37c),
                (27, 0xa9407212),
                (30, 0xfe8de0c9),
                (31, 0x00000007),
                ],
        },
    },
    Test {
        desc: "GTE_OP GTE_DCPL, lm=0, cv=0, v=0, mx=0, sf=1 random",
        initial: Config {
            controls: &[
                (0, 0xff35cdf4),
                (1, 0xf8acd6a6),
                (2, 0x1954aa70),
                (3, 0xae7b5062),
                (4, 0x00000c63),
                (5, 0xcad4cc39),
                (6, 0xb9c11958),
                (7, 0xa942b312),
                (8, 0xaf436779),
                (9, 0x3c2d507a),
                (10, 0x95f99741),
                (11, 0x72413224),
                (12, 0x0000499d),
                (13, 0x0a37d280),
                (14, 0xdbe8feec),
                (15, 0x2395909a),
                (16, 0x47364c98),
                (17, 0x795c2ed7),
                (18, 0x637e48f4),
                (19, 0x89557da5),
                (20, 0xffff997a),
                (21, 0x690eb551),
                (22, 0x3dfb368e),
                (23, 0x2bbe355f),
                (24, 0xb07c9d22),
                (25, 0x030c876b),
                (26, 0x00003b7d),
                (27, 0x0000765a),
                (28, 0x228c2901),
                (29, 0xffffe86f),
                (30, 0xffffaf93),
                (31, 0xc741f000),
                ],
            data: &[
                (0, 0x91d5c574),
                (1, 0xffffdf9c),
                (2, 0xcea213bc),
                (3, 0x0000143e),
                (4, 0x2360a947),
                (5, 0x00003248),
                (6, 0x1747e72e),
                (7, 0x0000cc08),
                (8, 0x0000381d),
                (9, 0xffffe2ff),
                (10, 0xffffe0f8),
                (11, 0xffffe1b6),
                (12, 0x9da7438d),
                (13, 0xff60f0ed),
                (14, 0xbf5961ab),
                (15, 0xbf5961ab),
                (16, 0x0000b1c1),
                (17, 0x0000dda6),
                (18, 0x0000ce75),
                (19, 0x0000b2d1),
                (20, 0xdb01b77a),
                (21, 0x19cd28cd),
                (22, 0x1a75d97a),
                (23, 0xe91dc0ad),
                (24, 0x764e464f),
                (25, 0x4aa5a1e5),
                (26, 0x3b1a1977),
                (27, 0x39fb3f5f),
                (30, 0xfe8de0c9),
                (31, 0x00000007),
                ],
        },
        command: 0x00080029,
        result: Config {
            controls: &[
                (0, 0xff35cdf4),
                (1, 0xf8acd6a6),
                (2, 0x1954aa70),
                (3, 0xae7b5062),
                (4, 0x00000c63),
                (5, 0xcad4cc39),
                (6, 0xb9c11958),
                (7, 0xa942b312),
                (8, 0xaf436779),
                (9, 0x3c2d507a),
                (10, 0x95f99741),
                (11, 0x72413224),
                (12, 0x0000499d),
                (13, 0x0a37d280),
                (14, 0xdbe8feec),
                (15, 0x2395909a),
                (16, 0x47364c98),
                (17, 0x795c2ed7),
                (18, 0x637e48f4),
                (19, 0x89557da5),
                (20, 0xffff997a),
                (21, 0x690eb551),
                (22, 0x3dfb368e),
                (23, 0x2bbe355f),
                (24, 0xb07c9d22),
                (25, 0x030c876b),
                (26, 0x00003b7d),
                (27, 0x0000765a),
                (28, 0x228c2901),
                (29, 0xffffe86f),
                (30, 0xffffaf93),
                (31, 0x81f80000),
                ],
            data: &[
                (0, 0x91d5c574),
                (1, 0xffffdf9c),
                (2, 0xcea213bc),
                (3, 0x0000143e),
                (4, 0x2360a947),
                (5, 0x00003248),
                (6, 0x1747e72e),
                (7, 0x0000cc08),
                (8, 0x0000381d),
                (9, 0x00007fff),
                (10, 0x00007fff),
                (11, 0x00007fff),
                (12, 0x9da7438d),
                (13, 0xff60f0ed),
                (14, 0xbf5961ab),
                (15, 0xbf5961ab),
                (16, 0x0000b1c1),
                (17, 0x0000dda6),
                (18, 0x0000ce75),
                (19, 0x0000b2d1),
                (20, 0x19cd28cd),
                (21, 0x1a75d97a),
                (22, 0x17ffffff),
                (23, 0xe91dc0ad),
                (24, 0x764e464f),
                (25, 0x0001bbae),
                (26, 0x0001a4e4),
                (27, 0x0001b87d),
                (28, 0x00007fff),
                (29, 0x00007fff),
                (30, 0xfe8de0c9),
                (31, 0x00000007),
                ],
        },
    },
    Test {
        desc: "GTE_OP GTE_NCS, lm=0, cv=0, v=0, mx=0, sf=1 random",
        initial: Config {
            controls: &[
                (0, 0xff35cdf4),
                (1, 0xf8acd6a6),
                (2, 0x1954aa70),
                (3, 0xae7b5062),
                (4, 0x00000c63),
                (5, 0xcad4cc39),
                (6, 0xb9c11958),
                (7, 0xa942b312),
                (8, 0xaf436779),
                (9, 0x3c2d507a),
                (10, 0x95f99741),
                (11, 0x72413224),
                (12, 0x0000499d),
                (13, 0x0a37d280),
                (14, 0xdbe8feec),
                (15, 0x2395909a),
                (16, 0x47364c98),
                (17, 0x795c2ed7),
                (18, 0x637e48f4),
                (19, 0x89557da5),
                (20, 0xffff997a),
                (21, 0x690eb551),
                (22, 0x3dfb368e),
                (23, 0x2bbe355f),
                (24, 0x307c9d22),
                (25, 0x030c876b),
                (26, 0x00003b7d),
                (27, 0x0000765a),
                (28, 0x228c2901),
                (29, 0xffffe86f),
                (30, 0xffffaf93),
                (31, 0xc741f000),
                ],
            data: &[
                (0, 0x91d5c574),
                (1, 0xffffdf9c),
                (2, 0xcea213bc),
                (3, 0x0000143e),
                (4, 0x2360a947),
                (5, 0x00003248),
                (6, 0x1747e72e),
                (7, 0x0000cc08),
                (8, 0x0000381d),
                (9, 0xffffe2ff),
                (10, 0xffffe0f8),
                (11, 0xffffe1b6),
                (12, 0x9da7438d),
                (13, 0xff60f0ed),
                (14, 0xbf5961ab),
                (15, 0xbf5961ab),
                (16, 0x0000b1c1),
                (17, 0x0000dda6),
                (18, 0x0000ce75),
                (19, 0x0000b2d1),
                (20, 0xdb01b77a),
                (21, 0x19cd28cd),
                (22, 0x1a75d97a),
                (23, 0xe91dc0ad),
                (24, 0x764e464f),
                (25, 0x4aa5a1e5),
                (26, 0x3b1a1977),
                (27, 0x39fb3f5f),
                (30, 0xfe8de0c9),
                (31, 0x00000007),
                ],
        },
        command: 0x0008001e,
        result: Config {
            controls: &[
                (0, 0xff35cdf4),
                (1, 0xf8acd6a6),
                (2, 0x1954aa70),
                (3, 0xae7b5062),
                (4, 0x00000c63),
                (5, 0xcad4cc39),
                (6, 0xb9c11958),
                (7, 0xa942b312),
                (8, 0xaf436779),
                (9, 0x3c2d507a),
                (10, 0x95f99741),
                (11, 0x72413224),
                (12, 0x0000499d),
                (13, 0x0a37d280),
                (14, 0xdbe8feec),
                (15, 0x2395909a),
                (16, 0x47364c98),
                (17, 0x795c2ed7),
                (18, 0x637e48f4),
                (19, 0x89557da5),
                (20, 0xffff997a),
                (21, 0x690eb551),
                (22, 0x3dfb368e),
                (23, 0x2bbe355f),
                (24, 0x307c9d22),
                (25, 0x030c876b),
                (26, 0x00003b7d),
                (27, 0x0000765a),
                (28, 0x228c2901),
                (29, 0xffffe86f),
                (30, 0xffffaf93),
                (31, 0x81f80000),
                ],
            data: &[
                (0, 0x91d5c574),
                (1, 0xffffdf9c),
                (2, 0xcea213bc),
                (3, 0x0000143e),
                (4, 0x2360a947),
                (5, 0x00003248),
                (6, 0x1747e72e),
                (7, 0x0000cc08),
                (8, 0x0000381d),
                (9, 0x00007fff),
                (10, 0xffff8000),
                (11, 0x00007fff),
                (12, 0x9da7438d),
                (13, 0xff60f0ed),
                (14, 0xbf5961ab),
                (15, 0xbf5961ab),
                (16, 0x0000b1c1),
                (17, 0x0000dda6),
                (18, 0x0000ce75),
                (19, 0x0000b2d1),
                (20, 0x19cd28cd),
                (21, 0x1a75d97a),
                (22, 0x17ff00ff),
                (23, 0xe91dc0ad),
                (24, 0x764e464f),
                (25, 0x0a38da48),
                (26, 0xdbe897a7),
                (27, 0x23958063),
                (28, 0x00007c1f),
                (29, 0x00007c1f),
                (30, 0xfe8de0c9),
                (31, 0x00000007),
                ],
        },
    },
    Test {
        desc: "GTE_OP GTE_NCT, lm=0, cv=0, v=0, mx=0, sf=1 random",
        initial: Config {
            controls: &[
                (0, 0xff35cdf4),
                (1, 0xf8acd6a6),
                (2, 0x1954aa70),
                (3, 0xae7b5062),
                (4, 0x00000c63),
                (5, 0xcad4cc39),
                (6, 0xb9c11958),
                (7, 0xa942b312),
                (8, 0xaf436779),
                (9, 0x3c2d507a),
                (10, 0x95f99741),
                (11, 0x72413224),
                (12, 0x0000499d),
                (13, 0x0a37d280),
                (14, 0xdbe8feec),
                (15, 0x2395909a),
                (16, 0x47364c98),
                (17, 0x795c2ed7),
                (18, 0x637e48f4),
                (19, 0x89557da5),
                (20, 0xffff997a),
                (21, 0x690eb551),
                (22, 0x3dfb368e),
                (23, 0x2bbe355f),
                (24, 0x307c9d22),
                (25, 0x030c876b),
                (26, 0x00003b7d),
                (27, 0x0000765a),
                (28, 0x228c2901),
                (29, 0xffffe86f),
                (30, 0xffffaf93),
                (31, 0xc741f000),
                ],
            data: &[
                (0, 0x91d5c574),
                (1, 0xffffdf9c),
                (2, 0xcea213bc),
                (3, 0x0000143e),
                (4, 0x2360a947),
                (5, 0x00003248),
                (6, 0x1747e72e),
                (7, 0x0000cc08),
                (8, 0x0000381d),
                (9, 0xffffe2ff),
                (10, 0xffffe0f8),
                (11, 0xffffe1b6),
                (12, 0x9da7438d),
                (13, 0xff60f0ed),
                (14, 0xbf5961ab),
                (15, 0xbf5961ab),
                (16, 0x0000b1c1),
                (17, 0x0000dda6),
                (18, 0x0000ce75),
                (19, 0x0000b2d1),
                (20, 0xdb01b77a),
                (21, 0x19cd28cd),
                (22, 0x1a75d97a),
                (23, 0xe91dc0ad),
                (24, 0x764e464f),
                (25, 0x4aa5a1e5),
                (26, 0x3b1a1977),
                (27, 0x39fb3f5f),
                (30, 0xfe8de0c9),
                (31, 0x00000007),
                ],
        },
        command: 0x00080020,
        result: Config {
            controls: &[
                (0, 0xff35cdf4),
                (1, 0xf8acd6a6),
                (2, 0x1954aa70),
                (3, 0xae7b5062),
                (4, 0x00000c63),
                (5, 0xcad4cc39),
                (6, 0xb9c11958),
                (7, 0xa942b312),
                (8, 0xaf436779),
                (9, 0x3c2d507a),
                (10, 0x95f99741),
                (11, 0x72413224),
                (12, 0x0000499d),
                (13, 0x0a37d280),
                (14, 0xdbe8feec),
                (15, 0x2395909a),
                (16, 0x47364c98),
                (17, 0x795c2ed7),
                (18, 0x637e48f4),
                (19, 0x89557da5),
                (20, 0xffff997a),
                (21, 0x690eb551),
                (22, 0x3dfb368e),
                (23, 0x2bbe355f),
                (24, 0x307c9d22),
                (25, 0x030c876b),
                (26, 0x00003b7d),
                (27, 0x0000765a),
                (28, 0x228c2901),
                (29, 0xffffe86f),
                (30, 0xffffaf93),
                (31, 0x81f80000),
                ],
            data: &[
                (0, 0x91d5c574),
                (1, 0xffffdf9c),
                (2, 0xcea213bc),
                (3, 0x0000143e),
                (4, 0x2360a947),
                (5, 0x00003248),
                (6, 0x1747e72e),
                (7, 0x0000cc08),
                (8, 0x0000381d),
                (9, 0x00007fff),
                (10, 0xffff8000),
                (11, 0x00007fff),
                (12, 0x9da7438d),
                (13, 0xff60f0ed),
                (14, 0xbf5961ab),
                (15, 0xbf5961ab),
                (16, 0x0000b1c1),
                (17, 0x0000dda6),
                (18, 0x0000ce75),
                (19, 0x0000b2d1),
                (20, 0x17ff00ff),
                (21, 0x17ff00ff),
                (22, 0x17ff00ff),
                (23, 0xe91dc0ad),
                (24, 0x764e464f),
                (25, 0x0a34aac5),
                (26, 0xdbe60855),
                (27, 0x239224a0),
                (28, 0x00007c1f),
                (29, 0x00007c1f),
                (30, 0xfe8de0c9),
                (31, 0x00000007),
                ],
        },
    },
];
